# Constructor

`Constructor` is a Python library designed to facilitate the simulation and exploration of [Constructor Theory](https://www.constructortheory.org/), a novel framework that reimagines the fundamental principles of physics. Rather than focusing on the dynamics of how systems evolve over time, Constructor Theory emphasizes what transformations (or "tasks") are possible or impossible within the physical laws of the universe.

With `Constructor`, researchers and enthusiasts can model and simulate abstract "constructors"—entities capable of performing tasks on physical systems without undergoing any net change themselves. The library allows users to define tasks, substrates, and conditions under which transformations can occur, enabling the exploration of fundamental questions about the nature of possibility and impossibility in physics.

Based on the theory developed by [David Deutsch](https://www.daviddeutsch.org.uk/) and [Chiara Marletto](https://www.chiaramarletto.com/)

## **Key Features**

- **Constructor Modeling**: Define constructors, abstract machines that can perform specific tasks without being degraded by the process. Constructors are central to the framework and can be customized to fit a variety of scenarios.

- **Task Definition**: Represent physical transformations as tasks that can be performed on substrates. Tasks define both the initial and final states of a system, as well as the conditions required for their execution.

- **Substrate Representation**: Model substrates as physical systems with defined states and properties. Substrates are the objects upon which tasks operate, and their states can be modified by constructors.

- **Condition Constraints**: Implement conditions that must be met for tasks to be possible. These can involve the state or properties of substrates, environmental factors, or probabilistic elements.

- **Simulation Engine**: Run simulations that apply constructors to substrates according to the tasks defined. The engine determines which tasks are possible or impossible under the given conditions, allowing users to explore the boundaries of physical transformations.

- **Extensibility**: Easily extend the library with more complex substrates, probabilistic tasks, quantum operations, or environmental conditions. The modular design allows for flexibility and growth as your research progresses.

- **Visualization and Analysis**: (Planned feature) Tools for visualizing the results of simulations, including state transitions and the network of possible transformations.

## **Use Cases**

- **Theoretical Physics Research**: Explore fundamental questions about what can or cannot be done within the laws of physics, testing hypotheses related to information theory, thermodynamics, and quantum mechanics.

- **Educational Tools**: Provide students with a hands-on way to understand the principles of Constructor Theory, allowing them to simulate and visualize abstract concepts.

- **Interdisciplinary Applications**: Apply the principles of Constructor Theory to other fields, such as biology (evolution of life), computer science (computation theory), and engineering (design of resilient systems).

## **Installation**

```bash
pip install constructor-lib
```

## **Usage**

Take a look at a simple example or one of the sample implementations:

- [Information](./examples/information.ipynb)
- [Thermodynamics](./examples/thermodynamics.ipynb)
- [Quantum](./examples/quantum.ipynb)
- [Cellular Automata](./examples/cellular_automata.ipynb)
